import inspect
import traceback

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models, transaction
from django.utils.module_loading import import_string
from manager_utils import sync, ManagerUtilsManager
import six

from jsonfield import JSONField


@six.python_2_unicode_compatible
class SmartManager(models.Model):
    """
    Specifies the template, its associated model template class, and whether or not deletions should
    be managed by this model template.
    """
    # A unique identifier to load smart managers by name
    name = models.CharField(max_length=128, unique=True, null=True, default=None)

    # The loadable model template class that inherits BaseSmartManager
    smart_manager_class = models.CharField(max_length=128)

    # True if this model template also deletes its objects when elements of the template
    # (or the template itself) is deleted
    manages_deletions = models.BooleanField(default=True)

    # The primary object that this smart manager manages
    primary_obj_type = models.ForeignKey(ContentType, null=True)
    primary_obj_id = models.PositiveIntegerField(default=0)
    primary_obj = GenericForeignKey('primary_obj_type', 'primary_obj_id')

    # The template of the model(s) being managed
    template = JSONField()

    objects = ManagerUtilsManager()

    def __str__(self):
        return str(self.name)

    def clean(self):
        """
        Verify that the object can be built and the template class can be loaded. If any exception happens, raise
        a validation error.
        """
        try:
            smart_manager = import_string(self.smart_manager_class)(self.template)
            smart_manager.build()
        except Exception as e:
            raise ValidationError('{0} - {1}'.format(str(e), traceback.format_exc()))

    @transaction.atomic
    def save(self, *args, **kwargs):
        """
        Builds the objects managed by the template before saving the template.
        """
        super(SmartManager, self).save(*args, **kwargs)

        smart_manager = import_string(self.smart_manager_class)(self.template)
        primary_built_obj = smart_manager.build()

        # Do an update of the primary object type and id after it has been built. We use an update since
        # you can't call save in a save method. We may want to put this in post_save as well later.
        if primary_built_obj:
            self.primary_obj_type = ContentType.objects.get_for_model(primary_built_obj)
            self.primary_obj_id = primary_built_obj.id
            SmartManager.objects.filter(id=self.id).update(
                primary_obj_type=self.primary_obj_type, primary_obj_id=primary_built_obj.id)

        # Sync all of the objects from the built template
        sync(self.smartmanagerobject_set.all(), [
            SmartManagerObject(
                smart_manager=self,
                model_obj_id=built_obj.id,
                model_obj_type=ContentType.objects.get_for_model(built_obj, for_concrete_model=False),
            )
            for built_obj in smart_manager.built_objs
        ], ['smart_manager_id', 'model_obj_id', 'model_obj_type_id'])


class SmartManagerObject(models.Model):
    """
    Tracks objects associated with model templates. This model is useful when model templates
    are used to manage deletions.
    """
    # The model template that manages this object
    smart_manager = models.ForeignKey(SmartManager)

    # The generic foreign key to the object
    model_obj_type = models.ForeignKey(ContentType)
    model_obj_id = models.PositiveIntegerField()
    model_obj = GenericForeignKey('model_obj_type', 'model_obj_id', for_concrete_model=False)

    objects = ManagerUtilsManager()

    class Meta:
        unique_together = ('model_obj_type', 'model_obj_id')


class SmartModelMixin(object):
    """
    A mixin for django models that provides smart manager behavior. The functions provided for models are

    smart_upsert: Upsert an existing model with a provided smart manager class and template,
    smart_delete: Delete an existing model and flush its associated smart manager.
    """
    def _get_smart_manager(self):
        """
        Gets the smart manager associated with this object or None if it doesn't have one.
        """
        smo = SmartManagerObject.objects.get_or_none(
            model_obj_type=ContentType.objects.get_for_model(self, for_concrete_model=False),
            model_obj_id=self.id)
        return smo.smart_manager if smo else None

    def smart_upsert(self, sm_class, sm_template):
        """
        Upserts an existing model using a smart manager class and template.
        """
        if not self.id:
            raise ValueError('Cannot call smart_upsert on a non-persisted model')

        sm = self._get_smart_manager()
        return SmartManager.objects.upsert(
            id=sm.id if sm else None,
            updates={
                'template': sm_template,
                'smart_manager_class': '{0}.{1}'.format(inspect.getmodule(sm_class).__name__, sm_class.__name__),
            })[0]

    def smart_delete(self):
        """
        Deletes the object's smart manager. Note that the object is normally managed by the smart manager, so the object
        is often deleted as well.
        """
        sm = self._get_smart_manager()
        self.delete()
        if sm:
            sm.delete()


class SmartManagerMixin(object):
    """
    Provides additional "smart" functions for Django model managers, including:

    smart_create: Creates an object using a smart manager and returns the smart manager.
    """
    def smart_create(self, sm_class, sm_template):
        """
        Given a smart manager class and template, constructs the object using the smart manager,
        and returns the smart manager object.
        """
        sm_class_path = '{0}.{1}'.format(inspect.getmodule(sm_class).__name__, sm_class.__name__)
        return SmartManager.objects.create(smart_manager_class=sm_class_path, template=sm_template)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='SmartManager',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, verbose_name='ID', primary_key=True)),
                ('name', models.CharField(unique=True, null=True, max_length=128, default=None)),
                ('smart_manager_class', models.CharField(max_length=128)),
                ('manages_deletions', models.BooleanField(default=True)),
                ('primary_obj_id', models.PositiveIntegerField(default=0)),
                ('template', jsonfield.fields.JSONField()),
                ('primary_obj_type', models.ForeignKey(to='contenttypes.ContentType', null=True)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='SmartManagerObject',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, verbose_name='ID', primary_key=True)),
                ('model_obj_id', models.PositiveIntegerField()),
                ('model_obj_type', models.ForeignKey(to='contenttypes.ContentType')),
                ('smart_manager', models.ForeignKey(to='smart_manager.SmartManager')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name='smartmanagerobject',
            unique_together=set([('model_obj_type', 'model_obj_id')]),
        ),
    ]



from smart_manager import BaseSmartManager
from smart_manager.tests.models import UpsertModel


class UpsertSmartManager(BaseSmartManager):
    def build(self):
        return self.build_obj(UpsertModel, char_field=self._template['char_field'], updates={
            'int_field': self._template['int_field'],
        })


class UpsertModelListTemplate(BaseSmartManager):
    def build(self):
        for template in self._template:
            self.build_obj(UpsertModel, char_field=template['char_field'], updates={
                'int_field': template['int_field'],
            })


class DontDeleteUpsertSmartManager(BaseSmartManager):
    def build(self):
        return self.build_obj(UpsertModel, char_field=self._template['char_field'], updates={
            'int_field': self._template['int_field'],
        }, is_deletable=False)



from django.test import TestCase

from smart_manager import admin


class AdminTest(TestCase):
    def test_admin(self):
        """
        Verify the admin is registered properly.
        """
        assert(admin)



from django.test import TestCase
from mock import patch

from smart_manager.base import BaseSmartManager
from smart_manager.tests.smart_managers import UpsertSmartManager
from smart_manager.tests.models import UpsertModel


class BaseSmartManagerTest(TestCase):
    """
    Tests functionality in the base model template.
    """
    def test_template_deepcopy(self):
        """
        Tests that the template can be modified after and it is still intact in the template
        class.
        """
        template = {'hello': 'world'}
        smart_manager = BaseSmartManager(template)

        # Modify the input template
        template['hello'] = 'world2'

        # The template in the model template should remain the same
        self.assertEquals(smart_manager._template, {'hello': 'world'})

    def test_build_object(self):
        """
        Tests that the build_object function upserts the object and adds it to the built objects list.
        """
        smart_manager = BaseSmartManager({})
        smart_manager.build_obj(UpsertModel, char_field='hi', updates={
            'int_field': 1,
        })
        smart_manager.build_obj(UpsertModel, char_field='hi', updates={
            'int_field': 2,
        })

        # There should be one UpsertModel with an int_field of 2
        upsert_model = UpsertModel.objects.get()
        self.assertEquals(upsert_model.char_field, 'hi')
        self.assertEquals(upsert_model.int_field, 2)
        self.assertEquals(UpsertModel.objects.count(), 1)
        self.assertEquals(smart_manager.built_objs, set([upsert_model]))

    def test_build_using(self):
        """
        Tests building using another model template.
        """
        smart_manager = BaseSmartManager({})
        built_objs = smart_manager.build_using(UpsertSmartManager, {'int_field': 1, 'char_field': '2'})
        self.assertTrue(type(built_objs) in (list, tuple,))

        upsert_model = UpsertModel.objects.get()
        self.assertEquals(upsert_model.int_field, 1)
        self.assertEquals(upsert_model.char_field, '2')
        self.assertEquals(UpsertModel.objects.count(), 1)
        self.assertEquals(smart_manager.built_objs, set([upsert_model]))

    @patch('smart_manager.tests.smart_managers.UpsertSmartManager.build', spec_set=True)
    def test_multi_build_using(self, mock_build):
        """
        Hits the branch for .build returning a list or tuple
        :type mock_build: Mock
        """
        mock_build.return_value = ['one', 'two']

        smart_manager = BaseSmartManager({})
        built_objs = smart_manager.build_using(UpsertSmartManager, {'int_field': 1, 'char_field': '2'})
        self.assertTrue(type(built_objs) in (list, tuple,))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django.db.models.deletion
import smart_manager.models


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='CantCascadeModel',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
            ],
        ),
        migrations.CreateModel(
            name='RelModel',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
            ],
        ),
        migrations.CreateModel(
            name='UpsertModel',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('char_field', models.CharField(max_length=128)),
                ('int_field', models.IntegerField()),
            ],
            bases=(models.Model, smart_manager.models.SmartModelMixin),
        ),
        migrations.AddField(
            model_name='cantcascademodel',
            name='rel_model',
            field=models.ForeignKey(to='tests.RelModel', on_delete=django.db.models.deletion.PROTECT),
        ),
    ]



from django.contrib import admin
from smart_manager.models import SmartManager


class SmartManagerAdmin(admin.ModelAdmin):
    list_display = ('name', 'smart_manager_class', 'primary_obj_type', 'primary_obj_id')


admin.site.register(SmartManager, SmartManagerAdmin)



from django.db.models.signals import pre_delete
from django.dispatch import receiver

from smart_manager.models import SmartManagerObject


@receiver(pre_delete, sender=SmartManagerObject, dispatch_uid='delete_model_obj_on_smart_manager_object_delete')
def delete_model_obj_on_smart_manager_object_delete(sender, instance, **kwargs):
    """
    If the model template is managing deletions, delete all of the model objects associated with it before
    the template is deleted.
    """
    if instance.smart_manager.manages_deletions:
        try:
            instance.model_obj.delete()
        except:
            # The model object could have been deleteed. Its ctype could have been corrupted. It could
            # have also been trying to cascade delete a protected model. Ignore any deletion errors with
            # model objects.
            pass



from copy import deepcopy

from manager_utils import upsert


class BaseSmartManager(object):
    def __init__(self, template):
        self._template = deepcopy(template)
        self._built_objs = set()

    @property
    def built_objs(self):
        return self._built_objs

    def build_obj(self, model_class, is_deletable=True, updates=None, defaults=None, **kwargs):
        """
        Builds an object using the upsert function in manager utils. If the object can be deleted
        by the smart manager, it is added to the internal _built_objs list and returned.
        """
        built_obj = upsert(model_class.objects, updates=updates, defaults=defaults, **kwargs)[0]
        if is_deletable:
            self._built_objs |= set([built_obj])

        return built_obj

    def build_using(self, smart_manager_class, template):
        """
        Builds objects using another builder and a template. Adds the resulting built objects
        from that builder to the built objects of this builder.
        """
        smart_manager = smart_manager_class(template)
        built_objs = smart_manager.build()
        self._built_objs |= smart_manager.built_objs

        # make sure build objs is a list or tuple
        if type(built_objs) not in (list, tuple,):
            built_objs = [built_objs]

        return built_objs

    def build(self):
        """
        All builders must implement the build function, which returns the built object. All build
        functions must also maintain an interal list of built objects, which are accessed by
        self.built_objs.
        """
        raise NotImplementedError



from django.apps import AppConfig


class SmartManagerConfig(AppConfig):
    name = 'smart_manager'
    verbose_name = 'Django Smart Manager'

    def ready(self):
        import smart_manager.signal_handlers
        assert(smart_manager)



from django.db import models

from smart_manager.models import SmartManagerMixin, SmartModelMixin


class UpsertModelManager(models.Manager, SmartManagerMixin):
    pass


class UpsertModel(models.Model, SmartModelMixin):
    """
    A model for testing upserts.
    """
    char_field = models.CharField(max_length=128)
    int_field = models.IntegerField()

    objects = UpsertModelManager()


class RelModel(models.Model):
    pass


class CantCascadeModel(models.Model):
    rel_model = models.ForeignKey(RelModel, on_delete=models.PROTECT)



from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import IntegrityError
from django.test import TestCase, TransactionTestCase
from django_dynamic_fixture import G, N

from smart_manager.models import SmartManager, SmartManagerObject
from smart_manager.tests.models import UpsertModel, RelModel, CantCascadeModel
from smart_manager.tests.smart_managers import UpsertModelListTemplate


class SmartManagerMixinTest(TransactionTestCase):
    """
    Tests the SmartManagerMixin class.
    """
    def test_smart_create(self):
        """
        Tests smart_create.
        """
        sm = UpsertModel.objects.smart_create(
            UpsertModelListTemplate, [{
                'char_field': 'valid',
                'int_field': 1,
            }])
        self.assertIsNone(sm.name)
        self.assertIsNotNone(sm.id)
        self.assertEquals(sm.smart_manager_class, 'smart_manager.tests.smart_managers.UpsertModelListTemplate')
        self.assertEquals(UpsertModel.objects.count(), 1)
        self.assertTrue(UpsertModel.objects.filter(char_field='valid', int_field=1).exists())

    def test_smart_dup_create(self):
        """
        Tests smart_create with integrity errors.
        """
        UpsertModel.objects.smart_create(
            UpsertModelListTemplate, [{
                'char_field': 'valid',
                'int_field': 1,
            }])
        with self.assertRaises(IntegrityError):
            UpsertModel.objects.smart_create(
                UpsertModelListTemplate, [{
                    'char_field': 'valid',
                    'int_field': 1,
                }])


class SmartModelMixinTest(TestCase):
    """
    Tests functionality in the SmartModelMixin.
    """
    def test_smart_upsert_non_persisted(self):
        """
        Tests a smart upsert on a non persisted model.
        """
        with self.assertRaises(ValueError):
            UpsertModel().smart_upsert(UpsertModelListTemplate, [{
                'char_field': 'valid',
                'int_field': 1,
            }])

    def test_smart_upsert(self):
        """
        Tests a smart upsert
        """
        um = G(UpsertModel, char_field='valid', int_field=1)

        um.smart_upsert(UpsertModelListTemplate, [{
            'char_field': 'valid',
            'int_field': 1,
        }])
        self.assertEquals(SmartManager.objects.count(), 1)
        self.assertEquals(SmartManagerObject.objects.count(), 1)

        um.smart_upsert(UpsertModelListTemplate, [{
            'char_field': 'valid2',
            'int_field': 2,
        }])
        self.assertEquals(SmartManager.objects.count(), 1)
        self.assertEquals(SmartManagerObject.objects.count(), 1)

        um = UpsertModel.objects.get()
        self.assertEquals(um.char_field, 'valid2')
        self.assertEquals(um.int_field, 2)

    def test_smart_delete_no_smart_manager(self):
        """
        Tests smart_delete when there was no previous smart manager.
        """
        um = G(UpsertModel, char_field='valid', int_field=1)
        um.smart_delete()

        self.assertFalse(UpsertModel.objects.filter(id=um.id).exists())

    def test_smart_delete_w_smart_manager(self):
        """
        Tests smart_delete when a smart manager exists.
        """
        UpsertModel.objects.smart_create(UpsertModelListTemplate, [{
            'char_field': 'valid',
            'int_field': 1,
        }])
        um = UpsertModel.objects.get()
        um.smart_delete()

        self.assertFalse(UpsertModel.objects.exists())
        self.assertFalse(SmartManager.objects.exists())


class ValidationTest(TransactionTestCase):
    """
    Tests that validation works appropriately.
    """
    def test_multi_null(self):
        """
        Tests that multiple smart managers with a null name can be created.
        """
        SmartManager.objects.create(
            name=None, smart_manager_class='smart_manager.tests.smart_managers.UpsertModelListTemplate',
            template=[{
                'char_field': 'valid1',
                'int_field': 1
            }],
        )
        SmartManager.objects.create(
            name=None, smart_manager_class='smart_manager.tests.smart_managers.UpsertModelListTemplate',
            template=[{
                'char_field': 'valid2',
                'int_field': 1
            }],
        )

    def test_invalid_load_path(self):
        """
        Tests that a validation error is raised on an invalid class path for the smart manager.
        """
        smart_manager = N(
            SmartManager,
            manages_deletions=True,
            smart_manager_class='smart_manager.tests.smart_managers.InvalidModelTemplate',
            template={},
        )
        with self.assertRaises(ValidationError):
            smart_manager.clean()

    def test_invalid_template(self):
        """
        Tests that a validation error is raised on an invalid template
        """
        smart_manager = N(
            SmartManager,
            manages_deletions=True,
            smart_manager_class='smart_manager.tests.smart_managers.UpsertModelListTemplate',
            template={'invalid': 'invalid'},
        )
        with self.assertRaises(ValidationError):
            smart_manager.clean()

    def test_valid_template(self):
        """
        Tests that a validation error is not raised on a valid template
        """
        self.assertFalse(UpsertModel.objects.exists())
        smart_manager = N(
            SmartManager,
            manages_deletions=True,
            smart_manager_class='smart_manager.tests.smart_managers.UpsertModelListTemplate',
            template=[{
                'char_field': 'valid',
                'int_field': 1
            }],
        )
        smart_manager.clean()


class SmartManagerTest(TestCase):
    """
    Tests custom functionality in the SmartManager class.
    """
    def test_str(self):
        smart_manager = SmartManager(name='hi')
        self.assertEquals(str(smart_manager), 'hi')

    def test_undeletable_object(self):
        """
        Tests deleting elements from the template of a smart manager when the elements are not
        deletable.
        """
        smart_manager = G(
            SmartManager,
            manages_deletions=True,
            smart_manager_class='smart_manager.tests.smart_managers.DontDeleteUpsertSmartManager',
            template={
                'char_field': 'hi',
                'int_field': 1,
            },
        )

        self.assertEquals(UpsertModel.objects.count(), 1)
        self.assertTrue(UpsertModel.objects.filter(char_field='hi', int_field=1).exists())

        # All objects should still exist even after deleting them from the template and deleting the
        # smart manager itself
        smart_manager.template = {
            'char_field': 'hello',
            'int_field': 2,
        }
        smart_manager.save()

        self.assertEquals(UpsertModel.objects.count(), 2)
        self.assertTrue(UpsertModel.objects.filter(char_field='hi', int_field=1).exists())
        self.assertTrue(UpsertModel.objects.filter(char_field='hello', int_field=2).exists())

        smart_manager.delete()
        self.assertEquals(UpsertModel.objects.count(), 2)

    def test_primary_obj_set(self):
        """
        Tests that the primary object is set during saving.
        """
        smart_manager = G(
            SmartManager,
            manages_deletions=True,
            smart_manager_class='smart_manager.tests.smart_managers.UpsertSmartManager',
            template={
                'char_field': 'hi',
                'int_field': 1,
            },
        )

        self.assertEquals(UpsertModel.objects.count(), 1)
        self.assertTrue(UpsertModel.objects.filter(char_field='hi', int_field=1).exists())
        self.assertEquals(smart_manager.primary_obj_id, UpsertModel.objects.get().id)
        self.assertEquals(smart_manager.primary_obj_type, ContentType.objects.get_for_model(UpsertModel))

        # Refresh the smart manager and verify the primary object params were persisted
        smart_manager = SmartManager.objects.get(id=smart_manager.id)
        self.assertEquals(smart_manager.primary_obj_id, UpsertModel.objects.get().id)
        self.assertEquals(smart_manager.primary_obj_type, ContentType.objects.get_for_model(UpsertModel))

    def test_template_changes(self):
        """
        Tests chaning the template and resaving.
        """
        smart_manager = G(
            SmartManager,
            manages_deletions=True,
            smart_manager_class='smart_manager.tests.smart_managers.UpsertModelListTemplate',
            template=[{
                'char_field': 'hi',
                'int_field': 1,
            }, {
                'char_field': 'hello',
                'int_field': 2,
            }],
        )

        self.assertEquals(UpsertModel.objects.count(), 2)
        self.assertTrue(UpsertModel.objects.filter(char_field='hi', int_field=1).exists())
        self.assertTrue(UpsertModel.objects.filter(char_field='hello', int_field=2).exists())

        smart_manager.template = [{
            'char_field': 'hi',
            'int_field': 1,
        }]
        smart_manager.save()

        self.assertEquals(UpsertModel.objects.count(), 1)
        self.assertTrue(UpsertModel.objects.filter(char_field='hi', int_field=1).exists())

        smart_manager.delete()
        self.assertFalse(UpsertModel.objects.exists())

    def test_template_changes_no_deletions(self):
        """
        Tests chaning the template and resaving when the template model does not manage deletions.
        """
        smart_manager = G(
            SmartManager,
            manages_deletions=False,
            smart_manager_class='smart_manager.tests.smart_managers.UpsertModelListTemplate',
            template=[{
                'char_field': 'hi',
                'int_field': 1,
            }, {
                'char_field': 'hello',
                'int_field': 2,
            }],
        )

        self.assertEquals(UpsertModel.objects.count(), 2)
        self.assertTrue(UpsertModel.objects.filter(char_field='hi', int_field=1).exists())
        self.assertTrue(UpsertModel.objects.filter(char_field='hello', int_field=2).exists())

        smart_manager.template = [{
            'char_field': 'hi',
            'int_field': 1,
        }]
        smart_manager.save()

        self.assertEquals(UpsertModel.objects.count(), 2)
        self.assertTrue(UpsertModel.objects.filter(char_field='hi', int_field=1).exists())
        self.assertTrue(UpsertModel.objects.filter(char_field='hello', int_field=2).exists())

        smart_manager.delete()
        self.assertEquals(UpsertModel.objects.count(), 2)
        self.assertTrue(UpsertModel.objects.filter(char_field='hi', int_field=1).exists())
        self.assertTrue(UpsertModel.objects.filter(char_field='hello', int_field=2).exists())

    def test_build_multiple_distinct_objs_delete(self):
        smart_manager = G(
            SmartManager,
            manages_deletions=True,
            smart_manager_class='smart_manager.tests.smart_managers.UpsertModelListTemplate',
            template=[{
                'char_field': 'hi',
                'int_field': 1,
            }, {
                'char_field': 'hello',
                'int_field': 2,
            }],
        )
        # Simulate re-saving
        smart_manager.save()

        self.assertEquals(UpsertModel.objects.count(), 2)
        self.assertTrue(UpsertModel.objects.filter(char_field='hi', int_field=1).exists())
        self.assertTrue(UpsertModel.objects.filter(char_field='hello', int_field=2).exists())

        smart_manager.delete()
        self.assertFalse(UpsertModel.objects.exists())

    def test_build_multiple_same_objs(self):
        smart_manager = G(
            SmartManager,
            manages_deletions=True,
            smart_manager_class='smart_manager.tests.smart_managers.UpsertModelListTemplate',
            template=[{
                'char_field': 'hi',
                'int_field': 1,
            }, {
                'char_field': 'hi',
                'int_field': 1,
            }],
        )
        # Simulate re-saving
        smart_manager.save()

        self.assertEquals(UpsertModel.objects.count(), 1)
        self.assertTrue(UpsertModel.objects.filter(char_field='hi', int_field=1).exists())

        smart_manager.delete()
        self.assertFalse(UpsertModel.objects.exists())

    def test_deletion_protected_model(self):
        """
        Tests when a cascade delete cant occur becasue of a protected model.
        """
        rel_model = G(RelModel)
        G(CantCascadeModel, rel_model=rel_model)

        smart_manager = G(
            SmartManager,
            manages_deletions=True,
            smart_manager_class='smart_manager.tests.smart_managers.UpsertSmartManager',
            template={
                'char_field': 'hi',
                'int_field': 1,
            },
        )
        SmartManagerObject.objects.create(smart_manager=smart_manager, model_obj=rel_model)

        smart_manager.delete()

        # The rel model should still exist since its protected. The smart manager object
        # should not
        self.assertFalse(SmartManagerObject.objects.exists())
        self.assertTrue(RelModel.objects.exists())

    def test_deletion_valid_obj_deletions_set(self):
        """
        Tests deletion of a valid object when deletions are set.
        """
        smart_manager = G(
            SmartManager,
            manages_deletions=True,
            smart_manager_class='smart_manager.tests.smart_managers.UpsertSmartManager',
            template={
                'char_field': 'hi',
                'int_field': 1,
            },
        )
        # Simulate re-saving
        smart_manager.save()

        upsert_model = UpsertModel.objects.get()
        self.assertEquals(upsert_model.char_field, 'hi')
        self.assertEquals(upsert_model.int_field, 1)

        # Delete the model template object, resulting in the deleting of the object it manages
        smart_manager.delete()
        self.assertEquals(UpsertModel.objects.count(), 0)

    def test_deletion_valid_obj_deletions_not_set(self):
        """
        Tests deletion of a valid object when deletions are not set.
        """
        smart_manager = G(
            SmartManager,
            manages_deletions=False,
            smart_manager_class='smart_manager.tests.smart_managers.UpsertSmartManager',
            template={
                'char_field': 'hi',
                'int_field': 1,
            },
        )
        # Simulate re-saving
        smart_manager.save()

        upsert_model = UpsertModel.objects.get()
        self.assertEquals(upsert_model.char_field, 'hi')
        self.assertEquals(upsert_model.int_field, 1)

        # Delete the model template object, resulting in the deleting of the object it manages
        smart_manager.delete()
        self.assertEquals(UpsertModel.objects.count(), 1)
